#include <math.h>
#include <stdio.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_statistics_float.h>

#include "libjpegrest.h"


/* All the functions se_* are for measuring the Squared Error statistic. */


/* Structures */
struct stat_struct
{
	int n;
	int stride;
	float *data;
};

/* Code */

/* Given an original and an approximation value,
   returns  Signal_Power/Noise_Power */
static float se_single(float original, float approximation)
{
	return (original - approximation) * (original - approximation);
}

/* Gets memory for a stat_struct, so a float for each pixel in a matrix
The matrices must be of same size, sure. */
static struct
stat_struct se_alloc(const gsl_matrix *m1, const gsl_matrix *m2)
{
	struct stat_struct tmp;

	tmp.data = NULL;

	if(m1->size1 != m2->size1 || m1->size2 != m2->size2)
	{
		fprintf(stderr, "Different sizes in se_get matrices.\n");
		return tmp;
	}

	tmp.n = m1->size1 * m1->size2;
	tmp.stride = 1;
	tmp.data = malloc(m1->size1 * m1->size2 * sizeof(float));

	return tmp;
}


/* Fills a stat_struct with the se of the pixel values of m1 and m2,
being m1 the original, and m2 the approximation
The matrices must be of same size, sure. */
static int
se_get(const gsl_matrix *m_orig, const gsl_matrix *m_approx,
	struct stat_struct se)
{
	int row, column;

	if(m_orig->size1 != m_approx->size1 || m_orig->size2 != m_approx->size2)
	{
		fprintf(stderr, "Different sizes in se_get matrices.\n");
		return -1;
	}

	for (row = 0; row < m_orig->size1; row++)
		for (column = 0; column < m_orig->size2; column++)
		{
			se.data[row * m_orig->size2 + column] =
				se_single(m_orig->data[row * m_orig->tda + column],
					m_approx->data[row * m_approx->tda + column]);
		}
	return 0;
}

/* Mean Squared Error */
static float
se_mse(const struct stat_struct s)
{
	return gsl_stats_float_mean(s.data, s.stride, s.n);
}

/* Variance of Squared Error */
static float
se_variance(const struct stat_struct s)
{
	return gsl_stats_float_variance(s.data, s.stride, s.n);
}

/* PSNR, as defined in "Lukasz Kizewski, Image Deblocking using local
segmentation, 2004".*/
static float
se_psnr(const struct stat_struct s)
{
	return 255. / sqrt(se_mse(s));
}

/* Free allocated statistical data */
static void
se_free(struct stat_struct s)
{
	free(s.data);
}

float
error_psnr(const gsl_matrix *m1, const gsl_matrix *m2)
{
	float psnr;
	struct stat_struct st;

	st = se_alloc(m1,m2);
	if (st.data == NULL)
		return -1;

	se_get(m1,m2,st);

	psnr = se_psnr(st);

	se_free(st);

	return psnr;
}

float
error_mse(const gsl_matrix *m1, const gsl_matrix *m2)
{
	float mse;
	struct stat_struct st;

	st = se_alloc(m1,m2);
	if (st.data == NULL)
		return -1;

	se_get(m1,m2,st);

	mse = se_mse(st);

	se_free(st);

	return mse;
}

float
error_se_variance(const gsl_matrix *m1, const gsl_matrix *m2)
{
	float variance;
	struct stat_struct st;

	st = se_alloc(m1,m2);
	if (st.data == NULL)
		return -1;

	se_get(m1,m2,st);

	variance = se_variance(st);

	se_free(st);

	return variance;
}
