#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "jdrest.h"

/* Implementation of the method developed by Mark A. Robertson and
 * Robert L. Stevenson:
 *   DCT Quantization Noise in Compressed Images
 * draft published online, dated of February 23rd, 2004.
 *
 * Notes:
 * - The article doesn't mention how to find a good \alpha^{(w)}.
 *   We start from an alpha equal to the proposed step size, and then
 *   halving until a good candidate is found (decreasing the term that we have
 *   to minimize). At MINIMUM_STEP_SIZE, we stop trying to minimize, and exit
 *   the minimizing loop.
 */

#define MAXSAMP 255.
#define MINSAMP 0.

#define MINIMUM_STEP_SIZE 0.0000001

LOCAL(void)
robertson_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
LOCAL(void)
project(struct jpeg_restoration *rest, RSAMPBUFFER in, RSAMPBUFFER out);
LOCAL(void)
min_map(struct jpeg_restoration *rest);
LOCAL(void)
get_new_estimate(struct jpeg_restoration *rest, float step_size);
LOCAL(float)
calc_objective(struct jpeg_restoration *rest, RSAMPBUFFER b);
LOCAL(void)
calc_gradient_hubber(struct jpeg_restoration *rest);
LOCAL(void)
add_noise_to_gradient(struct jpeg_restoration *rest);
LOCAL(float)
f_huber(struct jpeg_restoration *rest, float u);
LOCAL(float)
diff_huber(struct jpeg_restoration *rest, float u);
LOCAL(float)
diff2_huber(struct jpeg_restoration *rest, float u);
LOCAL(int)
inv_matrix(float *a);
LOCAL(void)
calc_inv_k_ey(struct jpeg_restoration *rest);


struct my_rest_struct
{
  struct jpeg_restoration pub;
  RSAMPBUFFER bitmap; /* Article's z. Final image for each step */
  RSAMPBUFFER first_bitmap; /* Article's z_q. Normally idct'ed picture. */
  RSAMPBUFFER estimate; /* Estimated new picture after each descent step */
  RSAMPBUFFER gradient; /* Gradient for the descent algorithm */
  RBLOCK qtable;
  RBLOCK inv_k_ey; /* DCT-domain error autocovariance */

  RCOEFBUFFER DCTmin, DCTmax, DCTimage;
  int width_in_blocks, height_in_blocks;
  int w, h; /* _blocks * 8 */

  /* Parameters of the algorithm */
  float step_size;
  int steps;
  int lambda;
  float T_Huber;
};

typedef struct my_rest_struct * my_rest_ptr;

/* Use jpeg_get_large(cinfo, size) */

METHODDEF(void) robertson_decompress_plane
  JPP((j_decompress_ptr cinfo, struct jpeg_restoration * rest,
    JSAMPARRAY output_buf, JBLOCKARRAY blockarray));

GLOBAL(struct jpeg_restoration *)
jpeg_rest_robertson(j_decompress_ptr cinfo, J_REST_TARGET target)
{
  my_rest_ptr rest;

  if (target != JREST_TARGET_IDCT_Y &&
    target != JREST_TARGET_IDCT_Cb &&
    target != JREST_TARGET_IDCT_Cr)
  {
    ERREXIT(cinfo, JERR_REST_NOT_APPLICABLE);
  }

  /* The first restoration */
  rest = (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
                                SIZEOF(struct my_rest_struct));

  rest->pub.decompress_plane = robertson_decompress_plane;
  rest->pub.target = target;
  add_restorator(cinfo, (struct jpeg_restoration *) rest);
  return (struct jpeg_restoration *) rest;
}

METHODDEF(void)
robertson_decompress_plane(j_decompress_ptr cinfo,
    struct jpeg_restoration * rest, JSAMPARRAY output_ptr,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,l;
  int buffer_row, block_rows;
  jpeg_component_info * compptr;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  JDIMENSION output_col;
  RSAMP value;

  compptr = &cinfo->comp_info[myrest->pub.component];

  if (cinfo->output_iMCU_row == 0)
  {
    robertson_restoration(cinfo, rest, blockarray);
  }

  /* Give the rows */

  buffer_row = cinfo->output_iMCU_row * compptr->v_samp_factor;
  
  /* How many rows we should give? */
  if (cinfo->output_iMCU_row < last_iMCU_row)
    block_rows = compptr->v_samp_factor;
  else {
    /* NB: can't use last_row_height here; it is input-side-dependent! */
    block_rows = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
    if (block_rows == 0) block_rows = compptr->v_samp_factor;
  }

  for (j=0; j<block_rows; j++)
  {
    output_col = 0;
    for (i=0; i<compptr->width_in_blocks; i++)
    {
      for(k=0; k<8;k++)
        for(l=0; l<8;l++)
        {
          value = myrest->bitmap[((buffer_row + j)*8+k)*myrest->w + (i*8+l)]
	    + 0.5;
          if (value > MAXSAMP)
            (output_ptr[k] + output_col)[l] = MAXSAMP;
          else if (value < MINSAMP)
            (output_ptr[k] + output_col)[l] = MINSAMP;
          else
            (output_ptr[k] + output_col)[l] = value;
            
        }

      output_col += compptr->DCT_scaled_size;
    }
    output_ptr += compptr->DCT_scaled_size;
  }

}

LOCAL(void)
get_env_configuration(j_decompress_ptr cinfo, struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  char *env_var;

  myrest->steps = 2;
  myrest->T_Huber = 0.1;
  myrest->step_size = 1.;
  myrest->lambda = 0.00075;

#ifndef NO_GETENV
  env_var = getenv("JPEG_ROBERTSON_STEPS");
  if (env_var != NULL)
    sscanf(env_var, "%i", &myrest->steps);
  
  env_var = getenv("JPEG_ROBERTSON_STEPSIZE");
  if (env_var != NULL)
    sscanf(env_var, "%f", &myrest->step_size);

  env_var = getenv("JPEG_ROBERTSON_THUBER");
  if (env_var != NULL)
    sscanf(env_var, "%f", &myrest->T_Huber);

  env_var = getenv("JPEG_ROBERTSON_LAMBDA");
  if (env_var != NULL)
    sscanf(env_var, "%f", &myrest->lambda);

#endif
}

/* !!! Some global functions for setting the variables are needed.
jpeg_rest_robertson_set_steps(rest, int)
jpeg_rest_robertson_set_step_size(rest, float)
... */

LOCAL(void)
robertson_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  jpeg_component_info * compptr;

  compptr = &cinfo->comp_info[myrest->pub.component];

  /* Prepare dimensions */
  myrest->width_in_blocks = compptr->width_in_blocks;
  myrest->height_in_blocks = compptr->height_in_blocks;
  myrest->w = compptr->width_in_blocks * 8;
  myrest->h = compptr->height_in_blocks * 8;

  /* Load the coefficients, and prepare the DCT ranges */
  set_DCT_constraint(cinfo, rest, blockarray);

  /* Allocate the bitmap */
  myrest->bitmap =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);
  
  /* Allocate the first bitmap */
  myrest->first_bitmap =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Allocate for the gradient */
  myrest->gradient =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Allocate the estimated bitmap */
  myrest->estimate =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Decompress the first image */
  fdct_inverse_image(myrest->DCTimage, myrest->bitmap, myrest->w,
      myrest->h);

  get_env_configuration(cinfo, rest);

  /* Do the minimization */
  min_map(rest);
}

LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,row,col, adr;
  static const double aanscalefactor[DCTSIZE] = 
    {
      1.0, 1.387039845, 1.306562965, 1.175875602,
      1.0, 0.785694958, 0.541196100, 0.275899379
    };
  RCOEF *table;

  myrest->DCTmin =
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTmax = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTimage = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);

  if ( (myrest->DCTmin == NULL) || (myrest->DCTmax == NULL) ||
      (myrest->DCTimage == NULL) )
    ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 0);
  
  /* Get quantization table */
  for (i=0; i < DCTSIZE2; i++)
    myrest->qtable[i] = cinfo->comp_info[myrest->pub.component].quant_table->quantval[i];
  table = myrest->qtable;

 
  /* For float AA&N IDCT method, multipliers are equal to quantization
   * coefficients scaled by scalefactor[row]*scalefactor[col], where
   *   scalefactor[0] = 1
   *   scalefactor[k] = cos(k*PI/16) * sqrt(2)    for k=1..7
   */
  i = 0;
  for (row = 0; row < DCTSIZE; row++) 
    for (col = 0; col < DCTSIZE; col++) 
      {
        table[i]=table[i]*aanscalefactor[row] * aanscalefactor[col];
        i++;
      }
  
  for (j=0;j<myrest->height_in_blocks;j++)
    for (i=0;i<myrest->width_in_blocks;i++)
      for (k=0;k<DCTSIZE2;k++)
      {
        adr = (j*myrest->width_in_blocks + i)*64 + k;
        myrest->DCTmin[adr]=((RCOEF) blockarray[j][i][k]-0.5)*table[k];
        myrest->DCTmax[adr]=((RCOEF) blockarray[j][i][k]+0.5)*table[k];
        myrest->DCTimage[adr]=(myrest->DCTmin[adr]+myrest->DCTmax[adr])/2.;
      }
}

/* Main algorithm minimization */
LOCAL(void)
min_map(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int step;
  float obj, last_obj;
  float ssize;

  last_obj = calc_objective(rest, myrest->bitmap);

  /* Save a copy of the first bitmap. z_q in the article. */
  /* !! It should be in some API */
  memcpy(myrest->first_bitmap, myrest->bitmap,
      sizeof(RSAMP)*myrest->h*myrest->w);

  /* Calculate the error autovariance using the qtable */
  calc_inv_k_ey(rest);

  for (step=0; step< myrest->steps; step++)
  {
    calc_gradient_hubber(rest);
    add_noise_to_gradient(rest);
    
    /* First step size to try. In the article they don't explain how to
     * calculate a good step size. We get it as a parameter. */
    ssize = myrest->step_size;

    /* Get a new estimate (w), searching a proper step size. */
    do
    {
      get_new_estimate(rest, ssize);
      obj = calc_objective(rest, myrest->estimate);

      /* For the possible next loop */
      ssize = ssize / 2.;
      if (ssize < MINIMUM_STEP_SIZE)
	break;
    } while (obj >= last_obj);
    /* If that's true, the image cannot be improved anymore.
     * So exit, because myrest->bitmap is good enough. */
    if (ssize < MINIMUM_STEP_SIZE)
      break;

    /*
    fprintf(stderr, "w: last_obj: %f, obj = %f, stepsize = %f\n", 
	last_obj, obj, ssize);
	*/

    last_obj = obj;
    project(rest, myrest->estimate, myrest->bitmap);
  }
}

LOCAL(void)
calc_inv_k_ey(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i;

  for (i = 0; i < DCTSIZE2; i++)
  {
    /* each element is the inverse of the variance of the coefficient.
     * Uniform distribution: sigma_e = (q_{i+1} - q_{i})^2 / 12 */
    myrest->inv_k_ey[i] = 12 / (myrest->qtable[i] * myrest->qtable[i]);
  }
}

LOCAL(void)
get_new_estimate(struct jpeg_restoration *rest, float step_size)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i;

  for (i = 0; i < myrest->h * myrest->w; i++)
    myrest->estimate[i] = myrest->bitmap[i] - step_size * myrest->gradient[i];
}

LOCAL(float)
calc_objective(struct jpeg_restoration *rest, RSAMPBUFFER b)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int index_mn, index_kl;
  int bound_up, bound_down, bound_left, bound_right;
  int border_local;
  int row, col, wrow, wcol;
  float obj;
  enum { WINDOW_LOCAL = 3 };

  border_local = (WINDOW_LOCAL-1)/2;

  obj = 0.;

  for(row=0; row < myrest->h; row++) 
    for (col=0; col < myrest->w; col++)
    {
      /* This scope is N_{m,n} as in the article */

      bound_up = (row >= border_local) ? row - border_local: 0;
      bound_down = (row < myrest->h - border_local) ? row + border_local
	  : myrest->h - 1;
      bound_left = (col >= border_local) ? col - border_local: 0;
      bound_right = (col < myrest->w - border_local) ? col + border_local
	  : myrest->w - 1;

      index_mn = row * myrest->w + col;

      for (wrow = bound_up; wrow <= bound_down; wrow++)
	for (wcol = bound_left; wcol <= bound_right; wcol++)
	{
	  if (wrow == row && wcol == col)
	    continue;

	  /* This scope is a "c in C", as explained in the article. */
	  
	  index_kl = wrow * myrest->w + wcol;

	  /* Calculate for the gradient */
	  /* sum(c in C, rho_T(d_c^t * z) */
	  obj += f_huber(rest, b[index_mn] - b[index_kl]);
	}
    }

  return obj;
}

/* Using the article notation, it gets Nabla r(z) from myrest->gradient,
 * multiplies it by lambda, and adds the herein calculated Nabla s(z). */
LOCAL(void)
add_noise_to_gradient(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,x,y,adr;
  int h,w;
  RSAMP bbitmap[DCTSIZE2];
  RCOEF coefdct[DCTSIZE2];
  RSAMP * ptr;
  float lambda;
  
  /* Algorithm parameter */
  lambda = myrest->lambda;

  h = myrest->h;
  w = myrest->w;


  for (i=0; i<h; i+=DCTSIZE)
    for (j=0; j<w;j+=DCTSIZE) 
    {
      adr=(i*w+j);    
      ptr = bbitmap;
      /* Put (z - z_q) in the bbitmap buffer */
      for (y=0;y<DCTSIZE;y++)
	for (x=0;x<DCTSIZE;x++)
	  *(ptr++)=myrest->bitmap[adr+x+y*w] - myrest->first_bitmap[adr+x+y*w] +
	    + 128; /* RAWCENTER for dct */

      /* Get D * (z - z_q) */
      fdct_buffer(bbitmap, coefdct);

      /* Get K_ey^-1 * D * (z - z_q) */
      for (y=0;y<DCTSIZE2;y++)
      {
	/*
	fprintf(stderr,"coefdct: %f inv_key: %f z-zq = %f\n", coefdct[y],
	    myrest->inv_k_ey[y], bbitmap[y]);
	    */
	coefdct[y] *= myrest->inv_k_ey[y];
      }

      /* Get D^t * K_ey^-1 * D * (z - z_q) */
      fidct_buffer(coefdct, bbitmap);

      /* Put the result into myrest->gradient, taking into account that
       * there is Nabla r(z) */
      ptr = bbitmap;
      for (y=0;y<DCTSIZE;y++)
	for (x=0;x<DCTSIZE;x++)
	{
	  /* 128 is rawcenter for the DCT operations */
	  /*
	  fprintf(stderr, "contribution (%f/%f): %f\n",
	      lambda * myrest->gradient[adr+x+y*w],
	      (*ptr - 128), lambda* myrest->gradient[adr+x+y*w] / *ptr);
	      */
	  myrest->gradient[adr+x+y*w] += lambda
	    * myrest->gradient[adr+x+y*w] + *(ptr++) - 128;
	}
    }
}

/* It puts, using the article notation, Nabla r(z) into myrest->gradient */
LOCAL(void)
calc_gradient_hubber(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int index_mn, index_kl;
  int bound_up, bound_down, bound_left, bound_right;
  int border_local;
  int row, col, wrow, wcol;
  float huber;
  enum { WINDOW_LOCAL = 3 };

  rest_rsbuffer_zero(myrest->gradient, myrest->h * myrest->w);

  border_local = (WINDOW_LOCAL-1)/2;

  for(row=0; row < myrest->h; row++) 
    for (col=0; col < myrest->w; col++)
    {
      /* This scope is N_{m,n} as in the article */

      bound_up = (row >= border_local) ? row - border_local: 0;
      bound_down = (row < myrest->h - border_local) ? row + border_local
	  : myrest->h - 1;
      bound_left = (col >= border_local) ? col - border_local: 0;
      bound_right = (col < myrest->w - border_local) ? col + border_local
	  : myrest->w - 1;

      index_mn = row * myrest->w + col;

      for (wrow = bound_up; wrow <= bound_down; wrow++)
	for (wcol = bound_left; wcol <= bound_right; wcol++)
	{
	  if (wrow == row && wcol == col)
	    continue;

	  /* This scope is a "c in C", as explained in the article. */
	  
	  index_kl = wrow * myrest->w + wcol;

	  /* Calculate for the gradient 
	   *   g^(k) = sum(c in C, d_c * rho_T' (d_c^t * z^(k)) ) */
	  huber = diff_huber(rest, myrest->bitmap[index_mn]
	      - myrest->bitmap[index_kl]);
	  myrest->gradient[index_mn] += huber;
	  myrest->gradient[index_kl] -= huber;
	}
    }
}

LOCAL(float)
f_huber(struct jpeg_restoration *rest, float u)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  float T;
  float ret;

  T = myrest->T_Huber;
  
  if (u > 0.)
  {
    if (u > T)
      ret = T*T + 2. * T * (u - T);
    else
      ret = u*u;
  }
  else
  {
    if (u < -T)
      ret = T*T + 2. * T * (-u - T);
    else
      ret = u*u;
  }

  /*fprintf(stderr, "fhuber(%f) = %f\n", u, ret);*/

  return ret;
}

LOCAL(float)
diff_huber(struct jpeg_restoration *rest, float u)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  float T;
  float ret;

  T = myrest->T_Huber;
  
  if (u > 0.)
  {
    if (u > T)
      ret = 2. * T;
    else
      ret = 2. * u;
  }
  else
  {
    if (u < -T)
      ret = -2. * T;
    else
      ret = 2. * u;
  }

  return ret;
}

LOCAL(float)
diff2_huber(struct jpeg_restoration *rest, float u)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  float T;
  float ret;

  T = myrest->T_Huber;
  
  if (u > 0.)
  {
    if (u > T)
      ret = 0;
    else
      ret = 2.;
  }
  else
  {
    if (u < -T)
      ret = 0;
    else
      ret = 2;
  }

  return ret;
}

/* Projection on the constraint - by Froment */
LOCAL(void)
project(struct jpeg_restoration *rest, RSAMPBUFFER in, RSAMPBUFFER out)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;

  int j;
  fdct_image(myrest->DCTimage, in, myrest->w, myrest->h);
    
  for (j=0;j<myrest->h*myrest->w;j++)
    if (myrest->DCTimage[j] > myrest->DCTmax[j])
      myrest->DCTimage[j] = myrest->DCTmax[j];
    else if (myrest->DCTimage[j] < myrest->DCTmin[j])
      myrest->DCTimage[j] = myrest->DCTmin[j];

  fdct_inverse_image(myrest->DCTimage, out, myrest->w, myrest->h);

}

LOCAL(void)
swap(float *a, float *b)
{
  float tmp;
  
  tmp = *a;
  *a = *b;
  *b = tmp;
}

/* Inverse an 8x8 matrix.
 * Return 1 on success, 0 on error (singular matrix).
 *
 * The method is almost literally coppiedfrom "Numerical Recipes in C" -
 * Gauss-Jordan */
LOCAL(int)
inv_matrix(float *a)
{
  enum {N = 8};
  int index_r[N], index_c[N], ipiv[N];
  int i, icol, irow, j, k, l, ll;
  float big, dum, pivinv, temp;

  for(j=0; j<N; j++)
    ipiv[j] = 0;
  for(i=0; i<N; i++)
  {
    big = 0.;
    for(j=0; j<N; j++)
      if (ipiv[j] != 1)
	for (k=0; k<N; k++)
	{
	  if (ipiv[k] == 0)
	  {
	    if (fabs(a[j*N+k]) >= big)
	    {
	      big = fabs(a[j*N+k]);
	      irow = j;
	      icol = k;
	    }
	  }
	  else if (ipiv[k] > 1)
	    return 0; /* Singular Matrix */
	}
    ++(ipiv[icol]);
    /* We have the pivot element. */
    if (irow != icol)
      for (l=0; l<N; l++)
	swap(&a[irow*N + l], &a[icol*N + l]);

    index_r[i] = irow;
    index_c[i] = icol;

    if (a[icol*N + icol] == 0.0)
      return 0; /* Singular Matrix */
    
    pivinv = 1.0 / a[icol*N + icol];
    a[icol*N + icol] = 1.0;
    for (l=0; l<N; l++)
      a[icol*N + l] *= pivinv;

    /* Reduce the rows */
    for (ll=0; ll < N; ll++)
      if (ll != icol)
      {
	dum = a[ll*N + icol];
	a[ll*N + icol] = 0.0;
	for (l=0; l<N; l++)
	  a[ll*N + l] -= a[icol*N + l] * dum;
      }
  }
  /* End of the main loop. Now we unscramble the solution */
  for (l=N-1; l>=0; l--)
    if (index_r[l] != index_c[l])
      for (k=0; k<N; k++)
	swap(&a[k*N + index_r[l]], &a[k*N + index_c[l]]);

  /* Done */
  return 1;
}
