#define JPEG_INTERNALS
#include "jinclude.h"
#include "jpeglib.h"
#include "jdrest.h"

/* Implementation of the method developed by Thomas P. O'Rourke and
 * Robert L. Stevenson:
 *   Improved Image Decompression for Reduced Transform Coding Artifacts
 * from IEEE Transactions on Circuits and Systems for Video Technology,
 * vol. 5, No. 6, December 1995.
 *
 * There are a few changes over the original:
 * - The gradient is calculated as:
 *   g^(k) = - sum(c in C, rho_T' (d_c^t * z^(k)) * d_c)
 *   instead of
 *   g^(k) = sum(c in C, rho_T' (d_c^t * z^(k)) * d_c^t)
 * - The step size isn't calculated. It's tried from 1, and halving
 *   as much as needed for minimizing the objective function
 *      sum(c in C, rho_T(z_{m,n} - z_{k,l}))
 * - The step size tried has a minimum on halving (MINIMUM_STEP_SIZE). Having
 *   reached the minimum, no more steps are done, as they wouldn't improve
 *   the bitmap according to the method constraints.
 */

#define MAXSAMP 255.
#define MINSAMP 0.

#define MINIMUM_STEP_SIZE 0.0000001

LOCAL(void)
orourke_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray);
LOCAL(void)
project(struct jpeg_restoration *rest, RSAMPBUFFER in, RSAMPBUFFER out);
LOCAL(void)
min_map(struct jpeg_restoration *rest);
LOCAL(void)
get_new_estimate(struct jpeg_restoration *rest, float step_size);
LOCAL(float)
calc_objective(struct jpeg_restoration *rest, RSAMPBUFFER b);
LOCAL(void)
calc_gradient(struct jpeg_restoration *rest);
LOCAL(float)
f_huber(struct jpeg_restoration *rest, float u);
LOCAL(float)
diff_huber(struct jpeg_restoration *rest, float u);
LOCAL(float)
diff2_huber(struct jpeg_restoration *rest, float u);


struct my_rest_struct
{
  struct jpeg_restoration pub;
  RSAMPBUFFER bitmap;
  RSAMPBUFFER estimate;
  RSAMPBUFFER gradient;
  RBLOCK qtable;

  RCOEFBUFFER DCTmin, DCTmax, DCTimage;
  int width_in_blocks, height_in_blocks;
  int w, h; /* _blocks * 8 */

  /* Parameters of the algorithm */
  float step_size;
  int steps;
  float T_Huber;
};

typedef struct my_rest_struct * my_rest_ptr;

/* Use jpeg_get_large(cinfo, size) */

METHODDEF(void) orourke_decompress_plane
  JPP((j_decompress_ptr cinfo, struct jpeg_restoration * rest,
    JSAMPARRAY output_buf, JBLOCKARRAY blockarray));

GLOBAL(struct jpeg_restoration *)
jpeg_rest_orourke(j_decompress_ptr cinfo, J_REST_TARGET target)
{
  my_rest_ptr rest;

  if (target != JREST_TARGET_IDCT_Y &&
    target != JREST_TARGET_IDCT_Cb &&
    target != JREST_TARGET_IDCT_Cr)
  {
    ERREXIT(cinfo, JERR_REST_NOT_APPLICABLE);
  }

  /* The first restoration */
  rest = (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_IMAGE,
                                SIZEOF(struct my_rest_struct));

  rest->pub.decompress_plane = orourke_decompress_plane;
  rest->pub.target = target;
  add_restorator(cinfo, (struct jpeg_restoration *) rest);
  return (struct jpeg_restoration *) rest;
}

METHODDEF(void)
orourke_decompress_plane(j_decompress_ptr cinfo, struct jpeg_restoration * rest,
    JSAMPARRAY output_ptr, JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,l;
  int buffer_row, block_rows;
  jpeg_component_info * compptr;
  JDIMENSION last_iMCU_row = cinfo->total_iMCU_rows - 1;
  JDIMENSION output_col;
  RSAMP value;

  compptr = &cinfo->comp_info[myrest->pub.component];

  if (cinfo->output_iMCU_row == 0)
  {
    orourke_restoration(cinfo, rest, blockarray);
  }

  /* Give the rows */

  buffer_row = cinfo->output_iMCU_row * compptr->v_samp_factor;
  
  /* How many rows we should give? */
  if (cinfo->output_iMCU_row < last_iMCU_row)
    block_rows = compptr->v_samp_factor;
  else {
    /* NB: can't use last_row_height here; it is input-side-dependent! */
    block_rows = (int) (compptr->height_in_blocks % compptr->v_samp_factor);
    if (block_rows == 0) block_rows = compptr->v_samp_factor;
  }

  for (j=0; j<block_rows; j++)
  {
    output_col = 0;
    for (i=0; i<compptr->width_in_blocks; i++)
    {
      for(k=0; k<8;k++)
        for(l=0; l<8;l++)
        {
          value = myrest->bitmap[((buffer_row + j)*8+k)*myrest->w + (i*8+l)]
	    + 0.5;
          if (value > MAXSAMP)
            (output_ptr[k] + output_col)[l] = MAXSAMP;
          else if (value < MINSAMP)
            (output_ptr[k] + output_col)[l] = MINSAMP;
          else
            (output_ptr[k] + output_col)[l] = value;
            
        }

      output_col += compptr->DCT_scaled_size;
    }
    output_ptr += compptr->DCT_scaled_size;
  }

}

LOCAL(void)
get_env_configuration(j_decompress_ptr cinfo, struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  char *env_var;

  myrest->steps = 2;
  myrest->T_Huber = 0.1;
  myrest->step_size = 1.;

#ifndef NO_GETENV
  env_var = getenv("JPEG_OROURKE_STEPS");
  if (env_var != NULL)
    sscanf(env_var, "%i", &myrest->steps);
  
  env_var = getenv("JPEG_OROURKE_STEPSIZE");
  if (env_var != NULL)
    sscanf(env_var, "%f", &myrest->step_size);

  env_var = getenv("JPEG_OROURKE_THUBER");
  if (env_var != NULL)
    sscanf(env_var, "%f", &myrest->T_Huber);

#endif
}

/* !!! Some global functions for setting the variables are needed.
jpeg_rest_orourke_set_steps(rest, int)
jpeg_rest_orourke_set_step_size(rest, float)
... */

LOCAL(void)
orourke_restoration(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  jpeg_component_info * compptr;

  compptr = &cinfo->comp_info[myrest->pub.component];

  /* Prepare dimensions */
  myrest->width_in_blocks = compptr->width_in_blocks;
  myrest->height_in_blocks = compptr->height_in_blocks;
  myrest->w = compptr->width_in_blocks * 8;
  myrest->h = compptr->height_in_blocks * 8;

  /* Load the coefficients, and prepare the DCT ranges */
  set_DCT_constraint(cinfo, rest, blockarray);

  /* Allocate the bitmap */
  myrest->bitmap =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Allocate for the gradient */
  myrest->gradient =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Allocate the estimated bitmap */
  myrest->estimate =
    (RSAMPBUFFER) rest_rsbuffer_malloc(cinfo,myrest->h*myrest->w);

  /* Decompress the first image */
  fdct_inverse_image(myrest->DCTimage, myrest->bitmap, myrest->w,
      myrest->h);

  get_env_configuration(cinfo, rest);

  /* Do the minimization */
  min_map(rest);
}

LOCAL(void)
set_DCT_constraint(j_decompress_ptr cinfo, struct jpeg_restoration *rest,
    JBLOCKARRAY blockarray)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i,j,k,row,col, adr;
  static const double aanscalefactor[DCTSIZE] = 
    {
      1.0, 1.387039845, 1.306562965, 1.175875602,
      1.0, 0.785694958, 0.541196100, 0.275899379
    };
  RCOEF *table;

  myrest->DCTmin =
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTmax = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);
  myrest->DCTimage = 
    (RCOEFBUFFER) rest_rcbuffer_malloc(cinfo,myrest->h*myrest->w);

  if ( (myrest->DCTmin == NULL) || (myrest->DCTmax == NULL) ||
      (myrest->DCTimage == NULL) )
    ERREXIT1(cinfo, JERR_OUT_OF_MEMORY, 0);
  
  /* Get quantization table */
  for (i=0; i < DCTSIZE2; i++)
    myrest->qtable[i] = cinfo->comp_info[myrest->pub.component].quant_table->quantval[i];
  table = myrest->qtable;

 
  /* For float AA&N IDCT method, multipliers are equal to quantization
   * coefficients scaled by scalefactor[row]*scalefactor[col], where
   *   scalefactor[0] = 1
   *   scalefactor[k] = cos(k*PI/16) * sqrt(2)    for k=1..7
   */
  i = 0;
  for (row = 0; row < DCTSIZE; row++) 
    for (col = 0; col < DCTSIZE; col++) 
      {
        table[i]=table[i]*aanscalefactor[row] * aanscalefactor[col];
        i++;
      }
  
  for (j=0;j<myrest->height_in_blocks;j++)
    for (i=0;i<myrest->width_in_blocks;i++)
      for (k=0;k<DCTSIZE2;k++)
      {
        adr = (j*myrest->width_in_blocks + i)*64 + k;
        myrest->DCTmin[adr]=((RCOEF) blockarray[j][i][k]-0.5)*table[k];
        myrest->DCTmax[adr]=((RCOEF) blockarray[j][i][k]+0.5)*table[k];
        myrest->DCTimage[adr]=(myrest->DCTmin[adr]+myrest->DCTmax[adr])/2.;
      }
}

/* Main algorithm minimization */
LOCAL(void)
min_map(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int step;
  float obj, last_obj;
  float ssize;

  last_obj = calc_objective(rest, myrest->bitmap);

  for (step=0; step< myrest->steps; step++)
  {
    calc_gradient(rest);
    
    /* First step size to try. The article calculates that using the
     * gradient and the image, but the algorithm proposed is heavy.
     * We get it as a parameter. */
    ssize = myrest->step_size;

    /* Get a new estimate (w), searching a proper step size. */
    do
    {
      get_new_estimate(rest, ssize);
      obj = calc_objective(rest, myrest->estimate);

      /* For the possible next loop */
      ssize = ssize / 2.;
      if (ssize < MINIMUM_STEP_SIZE)
	break;
    } while (obj >= last_obj);
    /* If that's true, the image cannot be improved anymore.
     * So exit, because myrest->bitmap is good enough. */
    if (ssize < MINIMUM_STEP_SIZE)
      break;

    /*fprintf(stderr, "w: last_obj: %f, obj = %f, stepsize = %f\n", 
	last_obj, obj, ssize);*/

    last_obj = obj;
    project(rest, myrest->estimate, myrest->bitmap);
  }
}

LOCAL(void)
get_new_estimate(struct jpeg_restoration *rest, float step_size)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int i;

  for (i = 0; i < myrest->h * myrest->w; i++)
    myrest->estimate[i] = myrest->bitmap[i] + step_size * myrest->gradient[i];
}

LOCAL(float)
calc_objective(struct jpeg_restoration *rest, RSAMPBUFFER b)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int index_mn, index_kl;
  int bound_up, bound_down, bound_left, bound_right;
  int border_local;
  int row, col, wrow, wcol;
  float obj;
  enum { WINDOW_LOCAL = 3 };

  border_local = (WINDOW_LOCAL-1)/2;

  obj = 0.;

  for(row=0; row < myrest->h; row++) 
    for (col=0; col < myrest->w; col++)
    {
      /* This scope is N_{m,n} as in the article */

      bound_up = (row >= border_local) ? row - border_local: 0;
      bound_down = (row < myrest->h - border_local) ? row + border_local
	  : myrest->h - 1;
      bound_left = (col >= border_local) ? col - border_local: 0;
      bound_right = (col < myrest->w - border_local) ? col + border_local
	  : myrest->w - 1;

      index_mn = row * myrest->w + col;

      for (wrow = bound_up; wrow <= bound_down; wrow++)
	for (wcol = bound_left; wcol <= bound_right; wcol++)
	{
	  if (wrow == row && wcol == col)
	    continue;

	  /* This scope is a "c in C", as explained in the article. */
	  
	  index_kl = wrow * myrest->w + wcol;

	  /* Calculate for the gradient */
	  /* sum(c in C, rho_T(d_c^t * z) */
	  obj += f_huber(rest, b[index_mn] - b[index_kl]);
	}
    }

  return obj;
}

LOCAL(void)
calc_gradient(struct jpeg_restoration *rest)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  int index_mn, index_kl;
  int bound_up, bound_down, bound_left, bound_right;
  int border_local;
  int row, col, wrow, wcol;
  float huber;
  enum { WINDOW_LOCAL = 3 };

  rest_rsbuffer_zero(myrest->gradient, myrest->h * myrest->w);

  border_local = (WINDOW_LOCAL-1)/2;

  for(row=0; row < myrest->h; row++) 
    for (col=0; col < myrest->w; col++)
    {
      /* This scope is N_{m,n} as in the article */

      bound_up = (row >= border_local) ? row - border_local: 0;
      bound_down = (row < myrest->h - border_local) ? row + border_local
	  : myrest->h - 1;
      bound_left = (col >= border_local) ? col - border_local: 0;
      bound_right = (col < myrest->w - border_local) ? col + border_local
	  : myrest->w - 1;

      index_mn = row * myrest->w + col;

      for (wrow = bound_up; wrow <= bound_down; wrow++)
	for (wcol = bound_left; wcol <= bound_right; wcol++)
	{
	  if (wrow == row && wcol == col)
	    continue;

	  /* This scope is a "c in C", as explained in the article. */
	  
	  index_kl = wrow * myrest->w + wcol;

	  /* Calculate for the gradient */
	  /* We change the calculation of the gradient. The article says:
	   *   g^(k) = sum(c in C, rho_T' (d_c^t * z^(k)) * d_c^t)
	   * but we do the next: 
	   *   g^(k) = - sum(c in C, rho_T' (d_c^t * z^(k)) * d_c) */
	  huber = diff_huber(rest, myrest->bitmap[index_mn]
	      - myrest->bitmap[index_kl]);
	  myrest->gradient[index_mn] -= huber;
	  myrest->gradient[index_kl] += huber;
	}
    }
}

LOCAL(float)
f_huber(struct jpeg_restoration *rest, float u)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  float T;
  float ret;

  T = myrest->T_Huber;
  
  if (u > 0.)
  {
    if (u > T)
      ret = T*T + 2. * T * (u - T);
    else
      ret = u*u;
  }
  else
  {
    if (u < -T)
      ret = T*T + 2. * T * (-u - T);
    else
      ret = u*u;
  }

  /*fprintf(stderr, "fhuber(%f) = %f\n", u, ret);*/

  return ret;
}

LOCAL(float)
diff_huber(struct jpeg_restoration *rest, float u)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  float T;
  float ret;

  T = myrest->T_Huber;
  
  if (u > 0.)
  {
    if (u > T)
      ret = 2. * T;
    else
      ret = 2. * u;
  }
  else
  {
    if (u < -T)
      ret = -2. * T;
    else
      ret = 2. * u;
  }

  return ret;
}

LOCAL(float)
diff2_huber(struct jpeg_restoration *rest, float u)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;
  float T;
  float ret;

  T = myrest->T_Huber;
  
  if (u > 0.)
  {
    if (u > T)
      ret = 0;
    else
      ret = 2.;
  }
  else
  {
    if (u < -T)
      ret = 0;
    else
      ret = 2;
  }

  return ret;
}

/* Projection on the constraint - by Froment */
LOCAL(void)
project(struct jpeg_restoration *rest, RSAMPBUFFER in, RSAMPBUFFER out)
{
  my_rest_ptr myrest = (my_rest_ptr) rest;

  int j;
  fdct_image(myrest->DCTimage, in, myrest->w, myrest->h);
    
  for (j=0;j<myrest->h*myrest->w;j++)
    if (myrest->DCTimage[j] > myrest->DCTmax[j])
      myrest->DCTimage[j] = myrest->DCTmax[j];
    else if (myrest->DCTimage[j] < myrest->DCTmin[j])
      myrest->DCTimage[j] = myrest->DCTmin[j];

  fdct_inverse_image(myrest->DCTimage, out, myrest->w, myrest->h);

}
