#include <stdio.h>
#include <jpeglib.h>


int
read_jpeg_header(const char *filename,
	struct jpeg_decompress_struct * srcinfo)
{
	struct jpeg_error_mgr jsrcerr;
	FILE *input_file;
	int i; /* iterator over APP markers */

	input_file = fopen(filename, "r");
	if (input_file == NULL)
	{
		fprintf(stderr,"The file \"%s\" cannot be opened.\n",
			filename);
		return -1;
	}

	/* Initialize the error handler for the jpeg library */
	srcinfo->err = jpeg_std_error(&jsrcerr);
	
	jpeg_create_decompress(srcinfo);

	jpeg_stdio_src(srcinfo, input_file);

	/* Prepare our interest in all markers */
	jpeg_save_markers(srcinfo, JPEG_COM, 0xFFFF);
	for (i=0; i < 16; i++)
		jpeg_save_markers(srcinfo, JPEG_APP0 + i , 0xFFFF);

	/* Fill the structure according to the jpeg header */
	jpeg_read_header(srcinfo, TRUE);

	fclose(input_file);

	return 0;
}

static const char * string_jfif_density_unit(int unit)
{
	char *tmpstring;
	switch(unit)
	{
		case 0:
			tmpstring = "No units. X and Y specify the pixel "
					"aspect ratio.";
			break;
		case 1:
			tmpstring = "Dots per inch";
			break;
		case 2:
			tmpstring = "Dots per cm";
			break;
		default:
			tmpstring = "Unknown by this application";
	}

	return tmpstring;
}
static const char * string_color_space(const J_COLOR_SPACE colorspace)
{
	char *tmpstring;
	switch(colorspace)
	{
		case JCS_UNKNOWN:
			tmpstring = "Unknown by the lib or not specified";
			break;
		case JCS_GRAYSCALE:
			tmpstring = "Grayscale";
			break;
		case JCS_RGB:
			tmpstring = "RGB";
			break;
		case JCS_YCbCr:
			tmpstring = "YCbCr (YUV)";
			break;
		case JCS_CMYK:
			tmpstring = "CMYK";
			break;
		case JCS_YCCK:
			tmpstring = "YCbCrK";
			break;
		default:
			tmpstring = "Unknown by this application";
	}

	return tmpstring;
}

static void print_quantization_matrix( JQUANT_TBL table)
{
	int row,column;

	/* prior to v6a of libjpeg, it was in zigzag order.
	   from v6a on, it's in natural order. */
	for (row=0; row<DCTSIZE; ++row)
	{
		for (column=0; column < DCTSIZE; ++column)
		{
			printf("%3u ", table.quantval[row * DCTSIZE + column] );
		}
		printf("\n");
	}
}

static void
print_quantization_matrices(const struct jpeg_decompress_struct * srcinfo)
{
	int i; /* For rolling along quantization tables */

	for(i=0; i< NUM_QUANT_TBLS; ++i)
	{
		if (srcinfo->quant_tbl_ptrs[i] != NULL)
		{
			printf("Quantization table %i -\n", i);
			print_quantization_matrix( *srcinfo->quant_tbl_ptrs[i]);
		}
	}

}

const char *
get_component_name(const unsigned int id, const J_COLOR_SPACE space)
{
	char * tmpstring;

	switch(space)
	{
		case JCS_UNKNOWN:
			tmpstring = "Unknown";
			break;
		case JCS_GRAYSCALE:
			if (id == 1)
				tmpstring = "Y";
			else
				tmpstring = "Unknown";
			break;
		case JCS_RGB:
			switch(id)
			{
			case 1:
				tmpstring = "R";
				break;
			case 2:
				tmpstring = "G";
				break;
			case 3:
				tmpstring = "B";
				break;
			default:
				tmpstring = "Unknown";
			}
			break;
		case JCS_YCbCr:
			switch(id)
			{
			case 1:
				tmpstring = "Y";
				break;
			case 2:
				tmpstring = "Cb";
				break;
			case 3:
				tmpstring = "Cr";
				break;
			default:
				tmpstring = "Unknown";
			}
			break;
		case JCS_CMYK:
			switch(id)
			{
			case 1:
				tmpstring = "C";
				break;
			case 2:
				tmpstring = "M";
				break;
			case 3:
				tmpstring = "Y";
				break;
			case 4:
				tmpstring = "K";
				break;
			default:
				tmpstring = "Unknown";
			}
			break;
		case JCS_YCCK:
			switch(id)
			{
			case 1:
				tmpstring = "Y";
				break;
			case 2:
				tmpstring = "Cb";
				break;
			case 3:
				tmpstring = "Cr";
				break;
			case 4:
				tmpstring = "K";
				break;
			default:
				tmpstring = "Unknown";
			}
			break;
		default:
			tmpstring = "Unknown";
	}

	return tmpstring;
}

static void
print_component_info(const struct jpeg_decompress_struct * srcinfo)
{
	int i; /* iterator over component */

	if(srcinfo->comp_info == NULL)
		return;

	for (i=0; i < srcinfo->num_components; ++i)
	{
		printf("Component %i (ID: %i, Index: %i): %s\n", i,
				srcinfo->comp_info[i].component_id,
				srcinfo->comp_info[i].component_index,
				get_component_name(
					srcinfo->comp_info[i].component_id,
					srcinfo->jpeg_color_space));
		printf(" Quantization table: %i\n",
				srcinfo->comp_info[i].quant_tbl_no);
		printf(" H-V sampling factor: %i-%i\n",
				srcinfo->max_h_samp_factor /
				srcinfo->comp_info[i].h_samp_factor,
				srcinfo->max_v_samp_factor /
				srcinfo->comp_info[i].v_samp_factor);
	}
}

static void
print_markers(const struct jpeg_decompress_struct * srcinfo)
{
	jpeg_saved_marker_ptr i; /* iterator through markers */
	int ichar; /* iterator over data chars */
	unsigned char value;

	/* APP0 marker */
	if (srcinfo->saw_JFIF_marker)
	{
		printf("JPEG marker: APP0\n");
		printf(" JFIF version: %u.%u\n",
				srcinfo->JFIF_major_version,
				srcinfo->JFIF_minor_version);
		printf(" JFIF density unit: %s (%u)\n", 
			string_jfif_density_unit(srcinfo->density_unit),
			srcinfo->density_unit);
		printf(" JFIF X pixel density: %u\n", srcinfo->X_density);
		printf(" JFIF Y pixel density: %u\n", srcinfo->Y_density);
	}
	/* APP14 marker */
	if (srcinfo->saw_Adobe_marker)
	{
		printf("JPEG marker: APP14\n");
		printf(" Adobe transform: %u\n",
				srcinfo->Adobe_transform);
	}

	i = srcinfo->marker_list;
	while (i != NULL)
	{
		printf("JPEG marker: ");
		switch(i->marker)
		{
			case JPEG_APP0+1:
				printf("APP1 (EXIF)\n");
				break;
			case JPEG_COM:
				printf("COM\n");
				break;
			default:
				printf("APP%u\n", i->marker - JPEG_APP0);
		}
		printf(" Data length: %u", i->data_length);
		printf(" Data: \"");
		for(ichar = 0; ichar < i->data_length; ++ichar)
		{
			value = i->data[ichar];
			/* If the character is printable */
			if ((value > 32 && value < 128) || value == '\n')
				putchar(value);
			else
				putchar('?');
		}
		printf("\"\n");
		i = i->next;
	}
}

static void
print_jpeg_struct_info(const struct jpeg_decompress_struct * srcinfo)
{
	printf("Encoded Image width: %u\n", srcinfo->image_width);
	printf("Encoded Image height: %u\n", srcinfo->image_height);
	printf("Number of components: %u\n", srcinfo->num_components);
	printf("JPEG Color space: %s\n",
		string_color_space(srcinfo->jpeg_color_space));
	printf("Precision bits in image data: %i\n",
		srcinfo->data_precision);

	print_markers(srcinfo);
	print_quantization_matrices(srcinfo);
	print_component_info(srcinfo);
}

int
main(int argn, char *argv[])
{
	struct jpeg_decompress_struct srcinfo;

	if (argn < 2)
	{
		fprintf(stderr, "usage: %s <file.jpg>\n",
			argv[0]);
		return -1;
	}


	if(read_jpeg_header(argv[1], &srcinfo) == -1)
		return -1;
	print_jpeg_struct_info(&srcinfo);

	jpeg_destroy_decompress(&srcinfo);

	return 0;
}
